import matplotlib.pyplot as plt
import cv2
import numpy as np
import scipy as sp
from sklearn.preprocessing import MinMaxScaler


def cube_plot(cube_plot,size,title):
    '''Plots a cube or RGB composite and normalizes the values to 0-255. 
    Inputs:
    cube_plot = cube,
    size = size of the plot,
    title = title.
    
    Outputs:
    The plot of the cube.'''
    #Copying the cube
    cube_plot2=cube_plot[:,:,:].copy()
    cube_plot2.data=np.nan_to_num(cube_plot.data)
    x,y,z=cube_plot[:,:,:].shape
    #Scale values to 0-255
    scaled=[]
    scaler=MinMaxScaler(feature_range=(0, 255))
    for band in range(cube_plot.data.shape[0]):
    
        scaledRGB=scaler.fit_transform(cube_plot2.data[band,:,:])
        scaled.append(scaledRGB)
            
    scaleda=np.array(scaled)
    cube_plot.data=scaleda.reshape(x,y,z)
    #Plotting the cube
    plot=cube_plot.plot.imshow(aspect=cube_plot.shape[2]/cube_plot.shape[1], size=size,add_labels=False, robust=True)
    plt.title(title)
    return (plot)


def image_plot(image_input,size2,title):
    '''Plots a single band image with an "Spectral" colormap. 
    
    Inputs:
    image_input = single band image,
    size = the size fo the plot, 
    title = title.
    
    Outputs:
    Plot of the image.'''
    
    plot2=image_input.plot.imshow(aspect=image_input.shape[1]/image_input.shape[0], size=size2, robust=True,add_labels=False, cmap='Spectral')
    plt.title(title)
    return (plot2)


def plot_comparison (cube_plot1, cube_plot2, band1,band2, title1, title2):
    '''Plots two selected band of a cube or cubes, to compare them. 
    
    Inputs: 
    cube_plot1 = the first cube, 
    cube_plot2 = the second cube,
    band1 = the band of the first cube,
    band2 = the band of the second cube,
    title1 = the title of the first cube,
    title2 = the title of the second cube.
    
    Outputs:
    The comparing plot of both bands.'''
    
    plot1, ax=plt.subplots(1,2)
    ax[0].imshow(cube_plot1[band1,:,:])
    ax[0].set_title(title1)
    ax[1].imshow(cube_plot2[band2,:,:])
    ax[1].set_title(title2)
    return(plt.show(plot1))


def fourier_plot (initial_cube,band,percentage_width, percentage_height):
    '''Plot the steps of the Fourier filtering to check the results. This function is only for viewing, filter the cube use the homonimous function under filtration. 
    
    Inputs: 
    initial_cube = the cube, 
    band = the band to check, 
    percentage_width = the width of the filter in percentage,
    percentage_height = and height of the filter in percentage.
    
    Otput:
    Image to check the Fourier filtering.'''
    
    fouraster2=cv2.dft(initial_cube.data[band,:,:], flags=cv2.DFT_COMPLEX_OUTPUT) 
    fouraster2_shift=np.fft.fftshift(fouraster2)
    #For plotting
    magnitude_spectrum = 20 * np.log((cv2.magnitude(fouraster2_shift[:, :, 0], fouraster2_shift[:, :, 1])))  
    y,z=initial_cube.data[band,:,:].shape
    
    rows, cols =initial_cube[band,:,:].shape
    mask = np.ones((rows, cols, 2), np.uint8)
    
    filter_width2=int((z*((100-percentage_width)/100))/2)
    filter_height2=int((y*(percentage_height/100))/2)

    #Mask
    cv2.rectangle(mask, (0,((y//2)-filter_height2)), (((z//2)-filter_width2),((y//2)+filter_height2)), 0, -1)
    cv2.rectangle(mask, (((z//2)+filter_width2),((y//2)-filter_height2)), (z,((y//2)+filter_height2)), 0, -1)

    #Mask applciation and reverse-fourier
    mfouraster2=fouraster2_shift*mask
    mshift_mask_mag = 20 * np.log(cv2.magnitude(mfouraster2[:, :, 0], mfouraster2[:, :, 1]))  #For plotting
    m_ishift2 = np.fft.ifftshift(mfouraster2)


    return_raster2=cv2.idft(m_ishift2)/(y*z)
    return_raster2= cv2.magnitude(return_raster2[:, :, 0], return_raster2[:, :, 1])
    
    fig = plt.figure(figsize=(10, 10))
    ax1 = fig.add_subplot(2,2,1)
    ax1.imshow(initial_cube.data[band,:,:], cmap="gray")
    ax1.title.set_text('Original Data')
    ax2 = fig.add_subplot(2,2,2)
    ax2.imshow(magnitude_spectrum, cmap="gray")
    ax2.title.set_text('Fourier of image')
    ax3 = fig.add_subplot(2,2,3)
    ax3.imshow(mshift_mask_mag, cmap="gray")
    ax3.title.set_text('Fourier + Mask')
    ax4 = fig.add_subplot(2,2,4)
    ax4.imshow(return_raster2.data, cmap="gray")
    ax4.title.set_text('After inverse Fourier')
    

def profiles_comparison(wavelengths,first_cube, second_cube,title1,title2, in_x, in_y,roi):
    '''Plot two spectral signatures to compare, the spectra is averaged in a window around the selected pixel. 
    Inputs:
    wavelengths =  the wavlengths, 
    first_cube = the first cube,
    second_cube = the second cube, 
    title1 = the title of the first cube,
    title2 = the title of the second cube, 
    in_x = the x position of the pixel to plot, 
    in_y = the y position of the pixel to plot
    roi = the size of the window in number of pixels (3 recommended).
    
    Outputs:
    The profiles comparison.'''
    
    #Defining the window
    stack_averaw=[]  
    roi_plus=int((roi/2)+0.5)  
    roi_minus=int((roi/2)-0.5)
    for band in range(first_cube.data.shape[0]):
        
        raw_cube2=first_cube[band,:,:]
        #Slicing the data to the ROI
        neighbourhood = raw_cube2[in_y-roi_minus:in_y+roi_plus, in_x-roi_minus:in_x+roi_plus] 
        #Average of the pixels
        average=np.mean(neighbourhood) 
        
        stack_averaw.append(average)
    first_average=np.array(stack_averaw)
    
    stack_avegauss=[]
    #Same for the filtered profile
    for band in range(second_cube.data.shape[0]):  
        
        gauss_cube2=second_cube[band,:,:]  
        neighbourhood2 = gauss_cube2[in_y-roi_minus:in_y+roi_plus, in_x-roi_minus:in_x+roi_plus]
        average2=np.mean(neighbourhood2)
        
        stack_avegauss.append(average2)
    second_average=np.array(stack_avegauss)
    
    plt.plot(wavelengths[0:len(first_average)], first_average, label=title1)
    plt.plot(wavelengths[0:len(second_average)],second_average, label=title2)
    plt.xlabel("Wavelengths (um)")
    plt.ylabel("Reflectance")
    plt.legend()
    plt.title("Profile comparison")
    return plt


def profile_plot (wavelengths,profile_singlecube,title_singleprofile, pixelx,pixely,roi):
    '''Plot the spectral signature of a cube, the spectra is averaged in a window around the selected pixel. 
    
    Inputs:
    wavelengths =  the wavelengths, 
    profile_singlecube = the first cube,
    title_singleprofile = the title of the first cube, 
    pixelx = the x position of the pixel to plot, 
    pixely = the y position of the pixel to plot
    roi = the size of the window in number of pixels (3 recommended).
    
    Outputs:
    The spectral profile.'''
    stack_ave=[]  
    roi_plus=int((roi/2)+0.5)  #Definign the window
    roi_minus=int((roi/2)-0.5)
    
    for band in range(profile_singlecube.data.shape[0]):
        
        neighbourhood = profile_singlecube[pixelx-roi_minus:pixelx+roi_plus, pixely-roi_minus:pixely+roi_plus]  #Slicing the data to the ROI
        average=np.mean(neighbourhood)  #Average of the pixels
        
        stack_ave.append(average)
    average2=np.array(stack_ave)
    
    fig=plt.plot(wavelenghts, average2)
    plt.title(title_singleprofile)
    return plt.show(fig)


def filter_comparison (cube_1,cube_2,title1,title2,band):
    '''Plots a comparison between the cubes before and after the filtration. It also plots the ratio between the cubes, and an image showing the pixels that changed more than 2% in black. 
    
    Inputs: 
    cube1 =  the cube before, 
    cube2 = the cube after,
    title1 = the title of te first one, 
    title2 = the title of the second one, 
    band = the band to check.
    
    Outputs:
    Plots to check the effect of the filtering.'''

    #Ratio of the two images
    ratio_cubes=cube_1/cube_2  
    #Calculating which pixels changed more than 2%
    cube_plus=cube_1+(cube_1*0.02)  
    cube_minus=cube_1-(cube_1*0.02)
    ratio_plus=cube_plus-cube_2
    ratio_minus=cube_2-cube_minus
    change_ratio=ratio_plus*ratio_minus
    change_ratio.data[change_ratio.data < 0]= 0
    #Plotting
    fig4, axs = plt.subplots(ncols=2,nrows=2, figsize=(10,10))  
    cube_1[band,:,:].plot.imshow(ax=axs[0,0],add_labels=False)
    axs[0,0].title.set_text(title1)
    axs[0,0].set_aspect(1)
    cube_2[band,:,:].plot.imshow(ax=axs[0,1],add_labels=False)
    axs[0,1].title.set_text(title2)
    axs[0,1].set_aspect(1)
    ratio_cubes[band,:,:].plot.imshow(ax=axs[1,0],robust=True,add_labels=False)
    axs[1,0].title.set_text('Ratio')
    axs[1,0].set_aspect(1)
    change_ratio[band,:,:].plot.imshow(ax=axs[1,1],robust=True,add_labels=False)
    axs[1,1].title.set_text('Change over 2%')
    axs[1,1].set_aspect(1)
    return 


def convexhull_plot(filtered_cube, wavelengths_full,mid_point,y_hull,x_hull):
    '''Plots the result of the convex hull continuum-removal method for a pixel. This function is only for viewing, to change the removal use the homonymous function under Preparation. 
    Inputs:
    filtered_cube = filtered cube,
    wavelengths_full = the wavelengths, 
    mid_point = the tie-point cube, 
    y_hull = the y position of the pixel, 
    x_hull = the x position of the pixel.
    
    Outputs:
    The plot of the continuum-removed spectrum (CH).'''
    
    wavelengths=wavelengths_full[0:76]
    average=filtered_cube[0:76,x_hull,y_hull]
    
    add_point=np.where(wavelengths==mid_point[x_hull,y_hull])[0]
    add_array=np.vstack((wavelengths[add_point], filtered_cube[add_point,x_hull,y_hull])).T
    #Temporarly removes the egdes of the spectrum
    points = np.c_[wavelengths, average]
    wavelengths, average = points.T
    augmented = np.concatenate([points, [(wavelengths[0], np.min(average)-1), (wavelengths[-1], np.min(average)-1)]], axis=0)
    #Performs the convex hull
    hull = sp.spatial.ConvexHull(augmented)
    pre_continuum_points = points[np.sort([v for v in hull.vertices if v < len(points)])]
    pre_continuum_points2 = np.concatenate((pre_continuum_points,add_array), axis=0)
    pre_continuum_points2.sort(axis=0)
    continuum_points=np.unique(pre_continuum_points2,axis=0)
    continuum_function = sp.interpolate.interp1d(*continuum_points.T)
    #Removes the continuum
    average_prime = average / continuum_function(wavelengths)
    average_prime[average_prime >= 1]= 1
    #Plot the image
    fig, axes = plt.subplots(2, 1, sharex=True)
    axes[0].plot(wavelengths, average, label='Data')
    axes[0].plot(*continuum_points.T, label='Continuum')
    axes[0].legend()
    axes[1].plot(wavelengths, average_prime, label='Data / Continuum')
    plt.xlabel("Wavelengths (um)")
    plt.ylabel("Reflectance")
    axes[1].legend()
    return plt


#second-and-first-order fit fit method
def safofit_plot(filtered_cube, removed_cube, wavelengths,y_plot,x_plot):    
    '''Plots the result of the second-and-first-order fit continuum-removal method for a pixel. The limits for the fits are manually defined using values established in the literature. This function is only for viewing, to perform the removal use the homonimous function under Preparation. 
    
    Inptus:
    filtered_cube = the filtered cube,
    removed_cube = the continuum-removed cube,
    y_plot = the y position of the pixel,
    x_plot = the x position of the pixel.
    
    Outputs:
    Plot of the continuum-removed spectrum (SAFO).'''

    #Second order fit for 1000 nm, it used a range for the two shoudlers around the 1000 nm absorption
    SAFO_cube=filtered_cube.data[0:74,x_plot,y_plot]  
    fitx10001=wavelengths[1:7]
    fitx10002=wavelengths[39:42]
    fitx1000=np.hstack((fitx10001,fitx10002))
    fity10001=SAFO_cube[1:7]
    fity10002=SAFO_cube[39:42]
    fity1000=np.hstack((fity10001,fity10002))
    fit1000=np.polyfit(fitx1000,fity1000,2)
    polival1000=np.polyval(fit1000,wavelengths[0:42])
    #Fit for 2000 nm, linear
    fitx2000=np.hstack((fitx10002,wavelengths[73])) 
    fity2000=np.hstack((fity10002,SAFO_cube[73]))
    fit2000=np.polyfit(fitx2000,fity2000,1)
    polival2000=np.polyval(fit2000,wavelengths[42:74])
    #Continuum removal by dividing
    continuum=np.hstack((polival1000,polival2000))  
    
    continum_removed=removed_cube[:,x_plot,y_plot]
    
    fig, axes = plt.subplots(2, 1, sharex=True)
    axes[0].plot(wavelengths[0:74], SAFO_cube, label='Data')
    axes[0].plot(wavelengths[0:74],continuum, label='Continuum')
    axes[0].legend()
    axes[1].plot(wavelengths[0:74], continum_removed[0:74], label='Data / Continuum')
    plt.xlabel("Wavelengths (um)")
    plt.ylabel("Reflectance")
    axes[1].legend()
    return plt


import numpy as np
import pysptools.spectro as spectro
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import scipy as sp
from joblib import Parallel, delayed
import multiprocessing


###DATA PRETARION

def attach_wave (initial_cube,wavelengths):
    '''This function eliminates the first two empty bands, turn all anomalous values to no data, and attach the wavelengths. 
    
    Inputs:
    initial_cube = M3 cube, 
    wavelengths = wavelengths file.
    
    Outputs:
    Prepared cube.'''

    #Removing bands
    initial_cube2=initial_cube[2:85,:,:]
    #Removing nodata
    initial_cube2.data[initial_cube2.data < -1000]= 0
    initial_cube2.data[initial_cube2.data > 1000]= 0
    #Attaching wavelength
    initial_cube2.coords['wavelength'] = ('band', wavelengths)
    cube_wave = initial_cube2.swap_dims({'band':'wavelength'})
    return cube_wave


def crop_cube (initial_cube,minnx,minny,maxxx,maxxy):
    '''Crop the prepared cube to a desired location, using the units of the coordinate system. Be caurefull to select values inside the coordinates of the image. 
    
    Inputs: 
    initial_cube = prepared cube, 
    minnx = lowest x limiting value,
    minny = lowest y limiting value,
    maxxx = highest x limiting value, 
    maxxy = highest y limiting value.
    
    Outputs:
    Cropped cube.'''

    #Check input coordinates
    if minnx < initial_cube.coords['x'].min().data or minny < initial_cube.coords['y'].min().data or maxxx > initial_cube.coords['x'].max().data or maxxy > initial_cube.coords['y'].max().data:
        raise ValueError("Invalid coordinate")
    #Assigning coordinates
    croped_cube=initial_cube.rio.clip_box(minx=minnx,miny=minny,maxx=maxxx,maxy=maxxy)
    #Creating the small rectangle
    rect_crop=patches.Rectangle((minnx,minny),(maxxx-minnx),(maxxy-minny),linewidth=1, edgecolor='r', facecolor='none')
    plot0, (axs, axs1)=plt.subplots(1,2, figsize=(5,20))
    #Plotting the cube
    initial_cube[0:3,:,:].plot.imshow(ax=axs,robust=True,add_labels=False)
    axs.add_patch(rect_crop)
    axs.set_aspect(1)
    axs.set_title('Full cube')
    #Plotting the cropped cube
    croped_cube[0:3,:,:].plot.imshow(ax=axs1,robust=True,add_labels=False)
    axs1.set_aspect(1)
    axs1.set_title('Cropped cube')
    plt.tight_layout() 
    return croped_cube

def crop_cube_size (initial_cube,cx1,cy1,cx2,cy2):
    '''Crop the prepared cube to a desired location, usign the number of lines and columns of the file. Be carefull to select values inside the image size. 
    
    Inputs: 
    initial_cube = prepared cube, 
    cx1 = lowest x limiting value,
    cy1 = highest x limiting value,
    cx2 = lowest y limiting value,
    cy2 = highest y limiting value.
    
    Outputs:
    Cropped cube.'''
    #Check input coordinates
    if cx1 < 0 or cy1 < 0 or cx2 > initial_cube.shape[2] or cy2 > initial_cube.shape[1]:
        raise ValueError("Invalid coordinate")
    M3_cubecrop=initial_cube[:,cy1:cy2,cx1:cx2]
    rect_crop=patches.Rectangle((cx1,cy1),(cx2-cx1),(cy2-cy1),linewidth=1, edgecolor='r', facecolor='none')

    plot0, ax=plt.subplots(1,2, figsize=(5,20))
    #To plot using the dimensions of the cube instead of the coordinates we should use only one band, in this case the 540 nm band
    ax[0].imshow(initial_cube[5,:,:], cmap='gray')
    ax[0].set_title('Full cube')
    ax[0].add_patch(rect_crop)
    ax[1].imshow(M3_cubecrop[5,:,:], cmap='gray')
    ax[1].set_title('Cropped cube')
    return M3_cubecrop

###CONTINUUM-REMOVAL

import multiprocessing
from joblib import Parallel, delayed

def midpoint(filtered_cube, wavelengths, peak_distance, peak_prominence, block_size=100, n_jobs=None):
    '''Finds the tie-point to limit the two absorption bands, used when the slope of the spectra is too steep. 
    It uses an automatic function to detect local peaks.
    
    Inputs:
    filtered_cube = filtered cube,
    wavelengths = wavelengths,
    peak_distance = the minimum distance between peaks (6 is recommended),
    peak_prominence = the minimum prominence of the peaks (0.002 is recommended).
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    Tie-points cube.'''

    x_midpoint, y_midpoint, z_midpoint = filtered_cube.shape
    midpoint_cube = filtered_cube[0, :, :].copy()

    # Resolve number of jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        cube_filtered = filtered_cube[20:60, a, b]
        if cube_filtered[39] == 0:
            return 0
        else:
            midpoint_y = np.array(cube_filtered[0])
            midpoint_y2 = np.append(midpoint_y, cube_filtered[39])
            midpoint_x = np.array(wavelengths[20])
            midpoint_x2 = np.append(midpoint_x, wavelengths[59])
            midpoint_fit = np.polyfit(midpoint_x2, midpoint_y2, 1)
            midpoint_polival = np.polyval(midpoint_fit, wavelengths[20:60])
            dif_cube = cube_filtered - midpoint_polival
            peaks, _ = sp.signal.find_peaks(dif_cube, distance=peak_distance, prominence=peak_prominence)
            if len(peaks) == 0:
                return wavelengths[42]
            else:
                return wavelengths[peaks + 20][-1]

    def process_block(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                block_result[a - r_start, b - c_start] = process_pixel(a, b)
        return block_result

    blocks = []
    for r in range(0, y_midpoint, block_size):
        for c in range(0, z_midpoint, block_size):
            r_end = min(r + block_size, y_midpoint)
            c_end = min(c + block_size, z_midpoint)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    full_array = np.zeros((y_midpoint, z_midpoint), dtype=np.float32)
    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        full_array[r_start:r_end, c_start:c_end] = results[idx]
        idx += 1

    midpoint_cube.data = full_array
    return midpoint_cube

##Convex hull method
 
def convexhull_removal(filtered_cube, wavelengths_full, mid_point, block_size=100, n_jobs=None):
    '''Remove the continuum of the spectra using the convex-hull method. 
    
    Inputs:
    filtered_cube = filtered cube, 
    wavelengths_full = wavelengths,
    mid_point = tie-point cube.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    Continuum removed cube by convex hull(CH).'''
    
    hull_cube = filtered_cube[:,:,:].copy()
    wavelengths = wavelengths_full
    bands, rows, cols = hull_cube.shape

    # Resolve number of jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        input_filtered = filtered_cube.data[:, a, b]
        input_midpoint = mid_point.data[a, b]

        if input_filtered[39] == 0:
            return np.zeros(bands, dtype=np.float32)

        try:
            add_point = np.where(wavelengths == input_midpoint)[0]
            add_array2 = np.vstack((wavelengths[add_point], input_filtered[add_point])).T
            points = np.c_[wavelengths, input_filtered]
            augmented = np.concatenate([
                points, 
                [(wavelengths[0], np.min(input_filtered) - 1), 
                 (wavelengths[-1], np.min(input_filtered) - 1)]
            ])
            hull = sp.spatial.ConvexHull(augmented)
            hull_vertices = [v for v in hull.vertices if v < len(points)]
            pre_continuum_points2 = points[np.sort(hull_vertices)]
            pre_continuum_points2 = np.concatenate((pre_continuum_points2, add_array2), axis=0)
            pre_continuum_points2 = np.unique(pre_continuum_points2, axis=0)
            pre_continuum_points2.sort(axis=0)
            continuum_function = sp.interpolate.interp1d(
                pre_continuum_points2[:, 0],
                pre_continuum_points2[:, 1],
                bounds_error=False,
                fill_value="extrapolate"
            )
            continuum = continuum_function(wavelengths)
            result = input_filtered / continuum
            result[result >= 1] = 1
            return np.nan_to_num(result).astype(np.float32)
        except:
            return np.zeros(bands, dtype=np.float32)

    def process_block(r_start, r_end, c_start, c_end):
        block_result = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                block_result.append(process_pixel(a, b))
        block_array = np.array(block_result, dtype=np.float32).reshape(r_end - r_start, c_end - c_start, bands)
        return block_array

    blocks = []
    for r in range(0, rows, block_size):
        for c in range(0, cols, block_size):
            r_end = min(r + block_size, rows)
            c_end = min(c + block_size, cols)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    full_array = np.zeros((bands, rows, cols), dtype=np.float32)
    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        block_array = results[idx]
        idx += 1
        full_array[:, r_start:r_end, c_start:c_end] = block_array.transpose(2, 0, 1)

    hull_cube.data = full_array
    return hull_cube

def find_minimums_ch(hull_cube, midpoint, wavelengths, block_size=100, n_jobs=None):
    '''This function finds the minimums around the 1 um and 2 um bands for the convex hull method. 
    
    Inputs:
    hull_cube = continuum-removed cube (CH),
    midpoint = tie-point,
    wavelengths = wavelengths.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    Minimum at 1 um and minimum at 2 um cubes.'''

    min1000 = hull_cube[0, :, :].copy()
    min2000 = hull_cube[0, :, :].copy()
    ymin, zmin = hull_cube[0, :, :].shape
    bands = hull_cube.shape[0]
    ofset = 5

    # Configurar núcleos para paralelización
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        # Default output: zeros
        val_min1000 = 0
        val_min2000 = 0
        input_hull = hull_cube.data[:, a, b]

        if input_hull[39] != 0:
            try:
                input_midpoint = midpoint.data[a, b]
                midpointp = int(np.where(wavelengths == input_midpoint)[0][0])

                # Minimum around 1 um band
                minimum_1000 = np.argmin(input_hull[0:midpointp])
                fitxp = max(0, minimum_1000 - ofset)
                fitxp2 = minimum_1000 + ofset
                if fitxp2 > midpointp:
                    fitxp2 = midpointp - 1
                fitx = wavelengths[fitxp:fitxp2]
                fity = input_hull[fitxp:fitxp2]

                fit_1000 = np.polyfit(fitx, fity, 2)
                polyval_1000 = np.polyval(fit_1000, wavelengths[fitxp:fitxp2])
                min1000p = np.where(polyval_1000 == np.min(polyval_1000))[0][0]
                final_1000 = wavelengths[min1000p + fitxp]

                if input_hull[min1000p + fitxp] < 0.974:
                    val_min1000 = final_1000
                else:
                    val_min1000 = 0

                # Minimum around 2 um band
                minimum_2000 = np.argmin(input_hull[midpointp:74]) + midpointp
                fit_start = max(0, minimum_2000 - ofset)
                fit_end = min(bands, minimum_2000 + ofset + 1)
                fitx2 = wavelengths[fit_start:fit_end]
                fity2 = input_hull[fit_start:fit_end]

                fit_2000 = np.polyfit(fitx2, fity2, 2)
                polyval_2000 = np.polyval(fit_2000, wavelengths[fit_start:fit_end])
                min2000p = np.where(polyval_2000 == np.min(polyval_2000))[0][0]

                final_index = min(max(min2000p + fit_start, midpointp + 1), 73)
                final_2000 = wavelengths[final_index]

                if input_hull[min2000p + fit_start] < 0.983:
                    val_min2000 = final_2000
                else:
                    val_min2000 = 0
            except:
                pass

        return val_min1000, val_min2000

    def process_block(r_start, r_end, c_start, c_end):
        min1000_block = []
        min2000_block = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                m1000, m2000 = process_pixel(a, b)
                min1000_block.append(m1000)
                min2000_block.append(m2000)
        shape = (r_end - r_start, c_end - c_start)
        return (np.array(min1000_block, dtype=np.float32).reshape(shape),
                np.array(min2000_block, dtype=np.float32).reshape(shape))

    blocks = []
    for r in range(0, ymin, block_size):
        for c in range(0, zmin, block_size):
            r_end = min(r + block_size, ymin)
            c_end = min(c + block_size, zmin)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    arr_min1000 = np.zeros((ymin, zmin), dtype=np.float32)
    arr_min2000 = np.zeros((ymin, zmin), dtype=np.float32)

    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        min1000_block, min2000_block = results[idx]
        arr_min1000[r_start:r_end, c_start:c_end] = min1000_block
        arr_min2000[r_start:r_end, c_start:c_end] = min2000_block
        idx += 1

    # Corrección de valores iguales a wavelengths[0]
    arr_min1000[arr_min1000 == wavelengths[0]] = wavelengths[18]

    min1000.data = arr_min1000
    min2000.data = arr_min2000

    return min1000, min2000 


def find_shoulders_ch(hull_cube, midpoint, min_1000, min_2000, wavelengths3, block_size=100, n_jobs=None):
    '''Find the shoulders around the minimums at 1 um and 2 um for the convex hull method. 
    
    Inputs:
    hull_cube = continuum removed cube (CH),
    midpoint = tie-point,
    min_1000 = the minimum at 1 um cube, 
    min_2000 = the minimum at 2 um cube,
    wavelengths3 = wavelengths.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    Left and right shoulders of the 1 um absorption band, left and right shoulder of the 2 um absorption band.'''

    # Crear cubos copia para los resultados
    shoulder0 = hull_cube[0, :, :].copy()
    shoulder1 = hull_cube[0, :, :].copy()
    shoulder2 = hull_cube[0, :, :].copy()
    shoulder3 = hull_cube[0, :, :].copy()
    y5, z5 = hull_cube[0, :, :].shape
    wavelengths = wavelengths3
    bands = hull_cube.shape[0]

    # Determinar núcleos para paralelización
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        result = [0, 0, 0, 0]  # [shoulder0, shoulder1, shoulder2, shoulder3]
        input_hull = hull_cube.data[:, a, b]

        try:
            input_mid = midpoint.data[a, b]
            midpoint_p = np.where(wavelengths == input_mid)[0][0]
        except:
            return result

        if min_1000[a, b] != 0:
            try:
                input_min1000 = min_1000.data[a, b]
                min1000_p = np.where(wavelengths == input_min1000)[0][0]
                ofset = 3

                # Left shoulder 1 μm
                shoulder_0 = np.where(input_hull[0:min1000_p] == max(input_hull[0:min1000_p]))[0][-1]
                fitxp0 = max(0, shoulder_0 - ofset)
                fitx0 = wavelengths[fitxp0:shoulder_0 + ofset]
                fity0 = input_hull[fitxp0:shoulder_0 + ofset]
                fit_0 = np.polyfit(fitx0, fity0, 2)
                polyval_0 = np.polyval(fit_0, wavelengths[fitxp0:shoulder_0 + ofset + 1])
                max0p = np.where(polyval_0 == max(polyval_0))[0][0]
                result[0] = wavelengths[max0p + fitxp0]

                # Right shoulder 1 μm
                shoulder_1 = np.where(input_hull[min1000_p:midpoint_p] == max(input_hull[min1000_p:midpoint_p]))[0][-1] + min1000_p
                fitxp1 = max(0, shoulder_1 - ofset)
                fitx1 = wavelengths[fitxp1:shoulder_1 + ofset]
                fity1 = input_hull[fitxp1:shoulder_1 + ofset]
                fit_1 = np.polyfit(fitx1, fity1, 2)
                polyval_1 = np.polyval(fit_1, wavelengths[fitxp1:shoulder_1 + ofset + 1])
                max1p = np.where(polyval_1 == max(polyval_1))[0][0]
                result[1] = wavelengths[max1p + fitxp1]
            except:
                pass

        if min_2000[a, b] != 0:
            try:
                input_min2000 = min_2000.data[a, b]
                min2000_p = np.where(wavelengths == input_min2000)[0][0]

                # Left shoulder 2 μm
                if midpoint_p - min2000_p < 0:
                    shoulder_2 = np.where(input_hull[midpoint_p:min2000_p] == max(input_hull[midpoint_p:min2000_p]))[0][-1] + midpoint_p
                    result[2] = wavelengths[shoulder_2]
                else:
                    result[2] = wavelengths[midpoint_p]

                # Right shoulder 2 μm
                shoulder_3 = np.where(input_hull[min2000_p:74] == max(input_hull[min2000_p:74]))[0][-1] + min2000_p
                result[3] = wavelengths[shoulder_3]
            except:
                pass

        return result

    def process_block(r_start, r_end, c_start, c_end):
        block_sh0 = []
        block_sh1 = []
        block_sh2 = []
        block_sh3 = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                s0, s1, s2, s3 = process_pixel(a, b)
                block_sh0.append(s0)
                block_sh1.append(s1)
                block_sh2.append(s2)
                block_sh3.append(s3)
        shape = (r_end - r_start, c_end - c_start)
        return (
            np.array(block_sh0, dtype=np.float32).reshape(shape),
            np.array(block_sh1, dtype=np.float32).reshape(shape),
            np.array(block_sh2, dtype=np.float32).reshape(shape),
            np.array(block_sh3, dtype=np.float32).reshape(shape)
        )

    # Crear bloques
    blocks = []
    for r in range(0, y5, block_size):
        for c in range(0, z5, block_size):
            r_end = min(r + block_size, y5)
            c_end = min(c + block_size, z5)
            blocks.append((r, r_end, c, c_end))

    # Paralelizar
    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    # Reconstruir los cubos
    shoulder0_arr = np.zeros((y5, z5), dtype=np.float32)
    shoulder1_arr = np.zeros((y5, z5), dtype=np.float32)
    shoulder2_arr = np.zeros((y5, z5), dtype=np.float32)
    shoulder3_arr = np.zeros((y5, z5), dtype=np.float32)

    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        sh0, sh1, sh2, sh3 = results[idx]
        shoulder0_arr[r_start:r_end, c_start:c_end] = sh0
        shoulder1_arr[r_start:r_end, c_start:c_end] = sh1
        shoulder2_arr[r_start:r_end, c_start:c_end] = sh2
        shoulder3_arr[r_start:r_end, c_start:c_end] = sh3
        idx += 1

    # Asignar a cubos originales
    shoulder0.data = shoulder0_arr
    shoulder1.data = shoulder1_arr
    shoulder2.data = shoulder2_arr
    shoulder3.data = shoulder3_arr

    return (shoulder0, shoulder1, shoulder2, shoulder3)

##second-and-first-order fit
def continuum_removal_SAFO(filtered_cube, wavelengths, order1, order2, block_size=100, n_jobs=None):
    '''Remove the continuum of the spectra using the second-and-first-order fit method. The limits for the fits are manually defined using values established
    in the literature.
    
    Inputs:
    filtered_cube = fitlered cube, 
    wavelengths = wavelengths,
    order1 = polynomial order for the first absorption band.
    order2 = polynomial order for the second absorption band.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).   
    
    Outputs:
    Continuum removed cube by second-and-first-order fit (SAFO).'''

    # Copy input cube to preserve original
    SAFO_cube = filtered_cube[:,:,:].copy()
    bands, rows, cols = SAFO_cube.shape

    # Resolve number of jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        # Extract spectrum for pixel (a, b)
        spectrum = filtered_cube.data[:, a, b]

        if spectrum[39] == 0:
            return np.zeros(83, dtype=np.float32)
        
        # Fit continuum at 1000 nm (second order by default)
        fitx1 = wavelengths[1:7]
        fitx2 = wavelengths[39:42]
        fity1 = spectrum[1:7]
        fity2 = spectrum[39:42]
        x_fit_1000 = np.hstack((fitx1, fitx2))
        y_fit_1000 = np.hstack((fity1, fity2))
        poly_1000 = np.polyfit(x_fit_1000, y_fit_1000, order1)
        continuum_1000 = np.polyval(poly_1000, wavelengths[0:42])

        # Fit continuum at 2000 nm (first order by default)
        x_fit_2000 = np.hstack((fitx2, wavelengths[73]))
        y_fit_2000 = np.hstack((fity2, spectrum[73]))
        poly_2000 = np.polyfit(x_fit_2000, y_fit_2000, order2)
        continuum_2000 = np.polyval(poly_2000, wavelengths[42:83])

        # Join the two parts of the continuum and apply the division
        continuum = np.hstack((continuum_1000, continuum_2000))
        result = spectrum / continuum
        result[result > 1] = 1

        return np.nan_to_num(result).astype(np.float32)

    def process_block(r_start, r_end, c_start, c_end):
        block_result = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                block_result.append(process_pixel(a, b))
        return np.array(block_result, dtype=np.float32).reshape(r_end - r_start, c_end - c_start, 83)

    # Define spatial blocks
    blocks = []
    for r in range(0, rows, block_size):
        for c in range(0, cols, block_size):
            r_end = min(r + block_size, rows)
            c_end = min(c + block_size, cols)
            blocks.append((r, r_end, c, c_end))

    # Process blocks in parallel
    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    # Reconstruct the cube
    full_array = np.zeros((83, rows, cols), dtype=np.float32)
    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        block = results[idx]
        idx += 1
        full_array[:, r_start:r_end, c_start:c_end] = block.transpose(2, 0, 1)

    SAFO_cube.data = full_array
    return SAFO_cube


def find_minimums_SAFO(SAFO_cube, wavelengths, block_size=100, n_jobs=None):
    '''This function finds the minimums around the 1 um and 2 um bands for the second-and-first-order fit method. 
    
    Inputs:
    SAFO_cube = continuum-removed cube (SAFO),
    wavelengths = wavelengths.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    Minimum at 1 um and minimum at 2 um cubes.'''


    min_1000SAFO = SAFO_cube[0,:,:].copy()
    min_2000SAFO = SAFO_cube[0,:,:].copy()
    y, z = SAFO_cube[0,:,:].shape
    bands = SAFO_cube.shape[0]
    ofsetSAFO = 5

    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        val_min_1000 = 0
        val_min_2000 = 0
        min_SAFO = SAFO_cube.data[:, a, b]

        if min_SAFO[39] != 0:
            try:
                # Minimum 1 um
                minimum_1000SAFO = np.argmin(min_SAFO[7:39]) + 7
                fitxpSAFO = max(0, minimum_1000SAFO - ofsetSAFO)
                fitxp2SAFO = minimum_1000SAFO + ofsetSAFO
                if fitxp2SAFO > 39:
                    fitxp2SAFO = 38
                fitxSAFO = wavelengths[int(fitxpSAFO):int(fitxp2SAFO)]
                fitySAFO = min_SAFO[int(fitxpSAFO):int(fitxp2SAFO)]
                fit_1000SAFO = np.polyfit(fitxSAFO, fitySAFO, 2)
                polyval_1000SAFO = np.polyval(fit_1000SAFO, wavelengths[int(fitxpSAFO):int(fitxp2SAFO)])
                min1000pSAFO = np.argmin(polyval_1000SAFO)
                final_1000SAFO = wavelengths[min1000pSAFO + int(fitxpSAFO)]

                if min_SAFO[min1000pSAFO + int(fitxpSAFO)] < 0.98:
                    val_min_1000 = final_1000SAFO
                else:
                    val_min_1000 = 0

                # Minimum 2 um
                minimum_2000SAFO = np.argmin(min_SAFO[39:74]) + 39
                min2000 = minimum_2000SAFO + ofsetSAFO
                if min2000 > 73:
                    min2000 = 73
                fitx2 = wavelengths[int(minimum_2000SAFO - ofsetSAFO):int(min2000)]
                fity2 = min_SAFO[int(minimum_2000SAFO - ofsetSAFO):int(min2000)]
                fit_2000SAFO = np.polyfit(fitx2, fity2, 2)
                polyval_2000SAFO = np.polyval(fit_2000SAFO, wavelengths[int(minimum_2000SAFO - ofsetSAFO):int(minimum_2000SAFO + ofsetSAFO + 1)])
                min2000pSAFO = np.argmin(polyval_2000SAFO)
                wave_index2000 = min2000pSAFO + minimum_2000SAFO - ofsetSAFO

                wave_index2000 = min(max(wave_index2000, 39), 73)
                final_2000SAFO = wavelengths[wave_index2000]

                if min_SAFO[wave_index2000] < 0.98:
                    val_min_2000 = final_2000SAFO
                else:
                    val_min_2000 = 0
            except:
                pass

        return val_min_1000, val_min_2000

    def process_block(r_start, r_end, c_start, c_end):
        min1000_block = []
        min2000_block = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                m1000, m2000 = process_pixel(a, b)
                min1000_block.append(m1000)
                min2000_block.append(m2000)
        shape = (r_end - r_start, c_end - c_start)
        return (np.array(min1000_block, dtype=np.float32).reshape(shape),
                np.array(min2000_block, dtype=np.float32).reshape(shape))

    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    arr_min_1000SAFO = np.zeros((y, z), dtype=np.float32)
    arr_min_2000SAFO = np.zeros((y, z), dtype=np.float32)

    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        min1000_block, min2000_block = results[idx]
        arr_min_1000SAFO[r_start:r_end, c_start:c_end] = min1000_block
        arr_min_2000SAFO[r_start:r_end, c_start:c_end] = min2000_block
        idx += 1

    # Corrección según el código original
    arr_min1000SAFOa = arr_min_1000SAFO
    arr_min1000SAFOa[arr_min1000SAFOa == wavelengths[0]] = wavelengths[18]

    min_1000SAFO.data = arr_min1000SAFOa
    min_2000SAFO.data = arr_min_2000SAFO

    return min_1000SAFO, min_2000SAFO


def find_shoulders_SAFO(SAFO_cube, min_1000SAFO, min_2000SAFO, wavelengths, block_size=100, n_jobs=None):
    '''Find the shoulders around the minimums at 1 um and 2 um for the second-and-first-order fit method. 
    
    Inputs:
    SAFO_cube = continuum removed cube (SAFO),
    min_1000SAFO = the minimum at 1 um cube, 
    min_2000SAFO = the minimum at 2 um cube,
    wavelengths = wavelengths.
    block_size= Size of the block during the paralelization.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).  
    
    Outputs:
    Left and right shoulders of the 1 um absorption band, left and right shoulder of the 2 um absorption band (the rigth shoulder of the 1 um is the same as the left shoulder of the 2 um absorption band).'''
    
    shoulder0SAFO = SAFO_cube[0, :, :].copy()
    shoulder1SAFO = SAFO_cube[0, :, :].copy()
    shoulder2SAFO = SAFO_cube[0, :, :].copy()
    y, z = SAFO_cube[0, :, :].shape
    bands = SAFO_cube.shape[0]
    ofsetSAFO = 3

    # Configurar núcleos para paralelización
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        s0, s1, s2 = 0, 0, 0
        input_shoulderSAFO = SAFO_cube.data[:, a, b]

        if min_1000SAFO[a, b] != 0:
            try:
                input_min1000SAFO = min_1000SAFO.data[a, b]
                min1000pSAFO = np.where(wavelengths == input_min1000SAFO)[0][0]
                shoulder_0SAFO = np.where(input_shoulderSAFO[0:min1000pSAFO] == max(input_shoulderSAFO[0:min1000pSAFO]))[0][-1]
                fitxp0SAFO = max(0, shoulder_0SAFO - ofsetSAFO)
                fitx0SAFO = wavelengths[fitxp0SAFO:shoulder_0SAFO + ofsetSAFO]
                fity0SAFO = input_shoulderSAFO[fitxp0SAFO:shoulder_0SAFO + ofsetSAFO]
                fit_0SAFO = np.polyfit(fitx0SAFO, fity0SAFO, 2)
                polyval_0SAFO = np.polyval(fit_0SAFO, wavelengths[fitxp0SAFO:shoulder_0SAFO + ofsetSAFO + 1])
                max0pSAFO = np.where(polyval_0SAFO == max(polyval_0SAFO))[0][0]
                s0 = wavelengths[max0pSAFO + fitxp0SAFO]
            except:
                pass

        if min_2000SAFO[a, b] != 0:
            try:
                input_min2000SAFO = min_2000SAFO.data[a, b]
                min2000pSAFO = np.where(wavelengths == input_min2000SAFO)[0][0]
                if min_1000SAFO[a, b] != 0:
                    min1000pSAFO = np.where(wavelengths == min_1000SAFO.data[a, b])[0][0]
                else:
                    min1000pSAFO = 0
                shoulder_1SAFO = np.where(
                    input_shoulderSAFO[min1000pSAFO:min2000pSAFO + 1] == max(input_shoulderSAFO[min1000pSAFO:min2000pSAFO + 1])
                )[0][-1] + min1000pSAFO
                fitxp1SAFO = max(0, shoulder_1SAFO - ofsetSAFO)
                maxs1 = min(shoulder_1SAFO + ofsetSAFO, bands - 1)
                fitx1SAFO = wavelengths[fitxp1SAFO:maxs1]
                fity1SAFO = input_shoulderSAFO[fitxp1SAFO:maxs1]
                fit_1SAFO = np.polyfit(fitx1SAFO, fity1SAFO, 2)
                polyval_1SAFO = np.polyval(fit_1SAFO, wavelengths[fitxp1SAFO:shoulder_1SAFO + ofsetSAFO + 1])
                max1pSAFO = np.where(polyval_1SAFO == max(polyval_1SAFO))[0][0]
                s1 = wavelengths[max1pSAFO + fitxp1SAFO]
                s2 = wavelengths[74]  # último valor
            except:
                pass

        return s0, s1, s2

    def process_block(r_start, r_end, c_start, c_end):
        sh0 = []
        sh1 = []
        sh2 = []
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                s0, s1, s2 = process_pixel(a, b)
                sh0.append(s0)
                sh1.append(s1)
                sh2.append(s2)
        shape = (r_end - r_start, c_end - c_start)
        return (
            np.array(sh0, dtype=np.float32).reshape(shape),
            np.array(sh1, dtype=np.float32).reshape(shape),
            np.array(sh2, dtype=np.float32).reshape(shape)
        )

    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    # Reconstrucción
    arr_sh0 = np.zeros((y, z), dtype=np.float32)
    arr_sh1 = np.zeros((y, z), dtype=np.float32)
    arr_sh2 = np.zeros((y, z), dtype=np.float32)

    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        sh0, sh1, sh2 = results[idx]
        arr_sh0[r_start:r_end, c_start:c_end] = sh0
        arr_sh1[r_start:r_end, c_start:c_end] = sh1
        arr_sh2[r_start:r_end, c_start:c_end] = sh2
        idx += 1

    shoulder0SAFO.data = arr_sh0
    shoulder1SAFO.data = arr_sh1
    shoulder2SAFO.data = arr_sh2

    return shoulder0SAFO, shoulder1SAFO, shoulder2SAFO



import numpy as np
import MoonIndex.preparation
import MoonIndex.filtration
import xarray as xa
from joblib import Parallel, delayed
import multiprocessing
import time

###CALCULATING ALL THE INDEXES

def indexes_total_CH(M3_cube,wavelengths, n_jobs=None):
    '''This function performs the full process of creating the indexes using the convex-hull removal method, from the filtering to the indexes generation. The attach_wave (cube_alone,wave) function must still be runned beforehand, but the user can input the full cube after that (it will take a long time), or crop it with crop_cube (initial_cube,minnx,minny,maxx,maxy) to save time. 
    
    Inputs:
    M3_cube = the cube, 
    wavelengths = the wavelengths.
    n_jobs= Caution. Amount of cores for paralelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores).If None, defaults to 1 (no parallelism).
    
    Outputs:
    An image with all the indexes processed (CH).'''
    start_total = time.time()

    def log_time(msg, t0):
        print(f"{msg}: {time.time() - t0:.2f} s")

    print("=== Starting the proccesing using CH ===")

    # Filtering
    t0 = time.time()
    fourier_cube = MoonIndex.filtration.fourier_filter(M3_cube, 60, 2)
    log_time("Fourier filtering", t0)

    t0 = time.time()
    gauss_cube = MoonIndex.filtration.gauss_filter(fourier_cube, wavelengths)
    log_time("Gaussian filtering", t0)

    # Continuum removal
    t0 = time.time()
    midpoint_cube = MoonIndex.preparation.midpoint(gauss_cube, wavelengths, 6, 0.002, block_size=100, n_jobs=n_jobs)
    log_time("Midpoint calculation", t0)

    t0 = time.time()
    hull_cube = MoonIndex.preparation.convexhull_removal(gauss_cube, wavelengths, midpoint_cube, block_size=100, n_jobs=n_jobs)
    log_time("Convex hull removal", t0)

    # Copy cube
    indexes_total = gauss_cube[0:28, :, :].copy()

    # Minima and shoulders
    t0 = time.time()
    M3_min1000ch, M3_min2000ch = MoonIndex.preparation.find_minimums_ch(hull_cube, midpoint_cube, wavelengths, block_size=100, n_jobs=n_jobs)
    log_time("Minimums calculation", t0)

    t0 = time.time()
    M3_shoulder0ch, M3_shoulder1ch, M3_shoulder2ch, M3_shoulder3ch = MoonIndex.preparation.find_shoulders_ch(
        hull_cube, midpoint_cube, M3_min1000ch, M3_min2000ch, wavelengths, block_size=100, n_jobs=n_jobs)
    log_time("Shoulders calculation", t0)

    # General indexes
    t0 = time.time()
    M3_R540 = R540(gauss_cube)
    M3_R1580 = R1580(gauss_cube)
    M3_sp = spinel(gauss_cube)
    M3_ol = olivine(gauss_cube)
    M3_cr = chromite(gauss_cube)
    M3_fe = iron(gauss_cube)
    M3_ti = titanium(gauss_cube)
    M3_clem = clementine(gauss_cube)
    M3_spanpx = RGB_spanpx(gauss_cube)
    log_time("General indexes", t0)

    # Convex hull indexes
    t0 = time.time()
    M3_BCI_CH = band_center(M3_min1000ch)
    M3_BCII_CH = band_center(M3_min2000ch)
    M3_BDI_CH = band_depth(hull_cube, M3_min1000ch, wavelengths,block_size=100, n_jobs=n_jobs)
    M3_BDII_CH = band_depth(hull_cube, M3_min2000ch, wavelengths,block_size=100, n_jobs=n_jobs)
    log_time("Band measurements", t0)

    t0 = time.time()
    M3_SSI_CH = SSI(gauss_cube, M3_shoulder1ch, wavelengths, block_size=100, n_jobs=n_jobs)
    M3_RGB8_CH = RGB8(gauss_cube, hull_cube, block_size=100, n_jobs=n_jobs)
    M3_RGB6_CH = RGB6(hull_cube)
    log_time("Others", t0)

    t0 = time.time()
    M3_BAI1000_CH = BA(hull_cube, wavelengths, M3_shoulder0ch, M3_shoulder1ch)
    M3_ASY1000_CH = ASY(hull_cube, wavelengths, M3_shoulder0ch, M3_shoulder1ch, M3_min1000ch,block_size=100, n_jobs=n_jobs)
    M3_BAI2000_CH = BA(hull_cube, wavelengths, M3_shoulder2ch, M3_shoulder3ch)
    M3_ASY2000_CH = ASY(hull_cube, wavelengths, M3_shoulder2ch, M3_shoulder3ch, M3_min2000ch,block_size=100, n_jobs=n_jobs)
    log_time("Area and asymmetry", t0)

    # Output stack
    t0 = time.time()
    indexes_total.data = np.dstack((
        M3_R540, M3_R1580, M3_sp, M3_ol, M3_cr, M3_fe, M3_ti,
        M3_clem[0], M3_clem[1], M3_clem[2],
        M3_spanpx[0], M3_spanpx[1], M3_spanpx[2],
        M3_BCI_CH, M3_BCII_CH, M3_BDI_CH, M3_BDII_CH, M3_SSI_CH,
        M3_RGB8_CH[0], M3_RGB8_CH[1], M3_RGB8_CH[2],
        M3_BAI1000_CH, M3_ASY1000_CH, M3_BAI2000_CH, M3_ASY2000_CH,
        M3_RGB6_CH[0], M3_RGB6_CH[1], M3_RGB6_CH[2]
    )).transpose(2, 0, 1)
    log_time("Stacking output bands", t0)

    # Naming bands
    bands = ['Reflectance 540 nm', 'Reflectance 1580 nm', 'Spinel parameter (Moriarty, 2022)',
             'Olivine parameter', 'Chromite parameter', 'Iron oxide parameter', 'Titanium parameter',
             'Clementine RED', 'Clementine GREEN', 'Clementine BLUE', 'Pyroxene parameter',
             'Spinel parameter (Pieters, 2014)', 'Anorthosite (Pieters, 2014)', 'Band center 1 µm CH',
             'Band center 2 µm CH', 'Band depth 1 µm CH', 'Band depth 2 µm CH', 'Spectral slope 1 µm CH',
             'Band depth 1.9 µm CH', 'Integrated band depth 2 µm CH', 'Integrated band depth 1 µm CH',
             'Band area 1 µm CH', 'Band assymetry 1 µm CH', 'Band area 2 µm CH', 'Band assymetry 2 µm CH',
             'Band depth at 950 nm CH', 'Band depth at 1.05 µm CH', 'Band depth at 1.25 µm CH']

    indexes_final_ch = xa.Dataset()
    for e in range(28):
        indexes_final_ch[bands[e]] = indexes_total[e, :, :]

    log_time("Total time", start_total)
    print("=== Finsihing calculation ===")

    return indexes_final_ch.astype(np.float32)


def indexes_total_SAFO(M3_cube, wavelengths, order1, order2, n_jobs=None):
    '''This function performs the full process of creating the indexes using the second-and-first-order fit removal method, from the filtering to the indexes generation. The attach_wave (cube_alone,wave) function must still be run beforehand, but the user can input the full cube after that (will take a long time), or crop it with crop_cube (initial_cube,minnx,minny,maxx,maxy) to save time. 
    
    Inputs:
    M3_cube = the cube, 
    wavelengths = the wavelengths,
    order1 = polynomial order for the first absorption band,
    order2 = polynomial order for the second absorption band.
    n_jobs= Caution. Amount of cores for parallelization. Use -1 for all cores, or a negative number to leave some cores 
    free (e.g., -2 means all but 2 cores). If None, defaults to 1 (no parallelism).
    
    Outputs:
    An image with all the indexes processed (SAFO).'''
    
    start_total = time.time()

    def log_time(msg, t0):
        print(f"{msg}: {time.time() - t0:.2f} s", flush=True)

    print("=== Starting the processing using SAFO ===", flush=True)

    # Filtering
    t0 = time.time()
    fourier_cube = MoonIndex.filtration.fourier_filter(M3_cube, 60, 2)
    gauss_cube = MoonIndex.filtration.gauss_filter(fourier_cube, wavelengths)  
    log_time("Filtering", t0)

    # Continuum removal
    t0 = time.time()
    SAFO_cube = MoonIndex.preparation.continuum_removal_SAFO(
        gauss_cube, wavelengths, order1, order2, block_size=100, n_jobs=n_jobs
    )
    log_time("Continuum removal (SAFO)", t0)

    # Copy to output cube
    indexes_total = gauss_cube[0:28, :, :].copy()

    # General indexes
    t0 = time.time()
    M3_R540 = R540(gauss_cube) 
    M3_R1580 = R1580(gauss_cube) 
    M3_sp = spinel(gauss_cube)  
    M3_ol = olivine(gauss_cube)  
    M3_cr = chromite(gauss_cube)
    M3_fe = iron(gauss_cube)
    M3_ti = titanium(gauss_cube)
    M3_clem = clementine(gauss_cube)
    M3_spanpx = RGB_spanpx(gauss_cube)
    log_time("General indexes", t0)

    # Find minimums and shoulders
    t0 = time.time()
    M3_min1000SAFO, M3_min2000SAFO = MoonIndex.preparation.find_minimums_SAFO(SAFO_cube, wavelengths, block_size=100, n_jobs=n_jobs)
    M3_shoulder0SAFO, M3_shoulder1SAFO, M3_shoulder2SAFO = MoonIndex.preparation.find_shoulders_SAFO(
        SAFO_cube, M3_min1000SAFO, M3_min2000SAFO, wavelengths, block_size=100, n_jobs=n_jobs)
    log_time("Minimums and shoulders", t0)

    # SAFO indexes
    t0 = time.time()
    M3_BCI_SAFO = band_center(M3_min1000SAFO)
    M3_BCII_SAFO = band_center(M3_min2000SAFO)
    M3_BDI_SAFO = band_depth(SAFO_cube, M3_min1000SAFO, wavelengths, block_size=100, n_jobs=n_jobs)
    M3_BDII_SAFO = band_depth(SAFO_cube, M3_min2000SAFO, wavelengths,block_size=100, n_jobs=n_jobs)
    log_time("Band measurements", t0)

    t0 = time.time()
    M3_SSI_SAFO = SSI(gauss_cube, M3_shoulder1SAFO, wavelengths, block_size=100, n_jobs=n_jobs)
    M3_RGB6_SAFO = RGB6(SAFO_cube)
    M3_RGB8_SAFO = RGB8(gauss_cube, SAFO_cube, block_size=100, n_jobs=n_jobs)
    log_time("Others", t0)

    t0 = time.time()
    M3_BAI1000_SAFO = BA(SAFO_cube, wavelengths, M3_shoulder0SAFO, M3_shoulder1SAFO)
    M3_ASY1000_SAFO = ASY(SAFO_cube, wavelengths, M3_shoulder0SAFO, M3_shoulder1SAFO, M3_min1000SAFO,block_size=100, n_jobs=n_jobs)
    M3_BAI2000_SAFO = BA(SAFO_cube, wavelengths, M3_shoulder1SAFO, M3_shoulder2SAFO)
    M3_ASY2000_SAFO = ASY(SAFO_cube, wavelengths, M3_shoulder1SAFO, M3_shoulder2SAFO, M3_min2000SAFO,block_size=100, n_jobs=n_jobs)
    log_time("Area and asymmetry", t0)

    # Output stack
    t0 = time.time()
    indexes_total.data = np.dstack((
        M3_R540, M3_R1580, M3_sp, M3_ol, M3_cr, M3_fe, M3_ti,
        M3_clem[0], M3_clem[1], M3_clem[2],
        M3_spanpx[0], M3_spanpx[1], M3_spanpx[2],
        M3_BCI_SAFO, M3_BCII_SAFO, M3_BDI_SAFO, M3_BDII_SAFO, M3_SSI_SAFO,
        M3_RGB8_SAFO[0], M3_RGB8_SAFO[1], M3_RGB8_SAFO[2],
        M3_BAI1000_SAFO, M3_ASY1000_SAFO, M3_BAI2000_SAFO, M3_ASY2000_SAFO,
        M3_RGB6_SAFO[0], M3_RGB6_SAFO[1], M3_RGB6_SAFO[2]
    )).transpose(2, 0, 1)
    log_time("Stacking output cube", t0)

    # Band naming
    t0 = time.time()
    bands = [
        'Reflectance 540 nm', 'Reflectance 1580 nm', 'Spinel parameter (Moriarty, 2022)',
        'Olivine parameter', 'Chromite parameter', 'Iron oxide parameter', 'Titanium parameter',
        'Clementine RED', 'Clementine GREEN', 'Clementine BLUE',
        'Pyroxene parameter', 'Spinel parameter (Pieters, 2014)', 'Anorthosite (Pieters, 2014)',
        'Band center 1 µm SAFO', 'Band center 2 µm SAFO',
        'Band depth 1 µm SAFO', 'Band depth 2 µm SAFO',
        'Spectral slope 1 µm SAFO',
        'Band depth 1.9 µm SAFO', 'Integrated band depth 2 µm SAFO', 'Integrated band depth 1 µm SAFO',
        'Band area 1 µm SAFO', 'Band asymmetry 1 µm SAFO', 'Band area 2 µm SAFO', 'Band asymmetry 2 µm SAFO',
        'Band depth at 950 nm SAFO', 'Band depth at 1.05 µm SAFO', 'Band depth at 1.25 µm SAFO'
    ]

    indexes_final_SAFO = xa.Dataset()
    for e in range(28):
        indexes_final_SAFO[bands[e]] = indexes_total[e, :, :]
    log_time("Dataset creation", t0)

    log_time("Total time SAFO", start_total)
    print("=== Finsihing calculation ===")
    return indexes_final_SAFO.astype(np.float32)

###INDIVIDUAL INDEXES

def R540 (gauss_cube):
    '''Creates the reflectance at 540 nm. 
    
    Input: 
    gauss_cube = the filtered cube.
    
    Output:
    R540 index.'''

    #The first band corresponds to that wavelength
    cube_R540=gauss_cube[0,:,:]  
    cube_R540.data[cube_R540.data==0]=np.nan
    return cube_R540


def R1580 (gauss_cube):
    '''Creates the reflectance at 1580 nm. 
    
    Input: 
    gauss_cube = the filtered cube.
    
    Output:
    R1580 index.'''

    #The first band corresponds to that wavelength
    cube_R1580=gauss_cube[47,:,:]  
    cube_R1580.data[cube_R1580.data==0]=np.nan
    return cube_R1580


def olivine (gauss_cube):
    '''Creates the olivine index. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output: 
    Olivine index.'''
    
    ol=(((gauss_cube[50,:,:]/((0.1*gauss_cube[21,:,:])+(0.1*gauss_cube[29,:,:])+(0.4*gauss_cube[35,:,:])+(0.4*gauss_cube[42,:,:])))-1))
    return ol


def spinel (gauss_cube):
    '''Creates the spinel index. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output: 
    Spinel index.'''
    
    sp1=((((gauss_cube[31,:,:]-gauss_cube[6,:,:])/500)*1350)+(gauss_cube[31,:,:]))/gauss_cube[73,:,:]
    return sp1


def chromite (gauss_cube):
    '''Creates the chromite index. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output: 
    Chromite index.'''
    
    cr=gauss_cube[0,:,:].copy()
    cr.data=((((gauss_cube[36,:,:]-gauss_cube[6,:,:])/600)*1500)+gauss_cube[36,:,:])/gauss_cube[77,:,:]
    return cr


def iron (gauss_cube):
    '''Creates the FeO index. Which return iron oxide abundance in percentage. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output: 
    FeO index.'''
    
    fe=gauss_cube[0,:,:].copy()
    fep=-np.arctan(((gauss_cube[16,:,:]/gauss_cube[6,:,:])-1.26)/(gauss_cube[6,:,]-0.01))
    fe.data=8.878 * (fep**1.8732)
    return fe


def titanium (gauss_cube):
    '''Creates the TiO index. Which returns titanium oxide abundance in percentage.
    
    Input:
    gauss_cube = filtered cube.
    
    Output: 
    TiO index.'''
    
    ti=gauss_cube[0,:,:].copy()
    tip=np.arctan(((gauss_cube[0,:,:]/gauss_cube[6,:,:])-0.45)/(gauss_cube[6,:,]-0.05))
    ti.data=2.6275 * (tip**4.2964)
    return ti


def clementine (gauss_cube):
    '''Creates the clementine-like index. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output:
    The clementine-like RGB composite.'''
    
    clem=gauss_cube[0:3,:,:].copy() 
    #Selecting the band to divide the filtered cube
    B1=gauss_cube[6,:,:]/gauss_cube[0,:,:]  
    B2=gauss_cube[6,:,:]/gauss_cube[19,:,:]
    B3=gauss_cube[0,:,:]/gauss_cube[6,:,:]
    clem.data=np.dstack((B1,B2,B3)).transpose(2,0,1)
    clem.data[clem.data > 3]=0
    return clem


def RGB_spanpx (gauss_cube):
    '''Creates the spanpx (spinel-anorthosite-pyroxene) index. 
    
    Input:
    gauss_cube = filtered cube.
    
    Output:
    The spanpx RGB composite.'''
    
    spanpx=gauss_cube[0:3,:,:].copy()  
    px=(gauss_cube[4,:,:]+gauss_cube[29,:,:])/gauss_cube[16,:,:]
    sp=gauss_cube[39,:,:]/gauss_cube[51,:,:]
    an=(gauss_cube[19,:,:]+gauss_cube[44,:,:])/gauss_cube[31,:,:]
    spanpx.data=np.dstack((px,sp,an)).transpose(2,0,1)
    spanpx.data[spanpx.data > 3]=0
    return spanpx


def band_center (minimum):
    '''Creates the band minimum, it works for both absorption bands by changing the corresponding inputs. 
    
    Input: 
    minimum = the minimum image.
    
    Output:
    The minimum of the selected abosprtion band.'''
    
    band_center=minimum.copy()
    band_center.data[band_center.data==0]=np.nan
    return band_center



def band_depth(removed_cube, minimum, wavelengths, block_size=100, n_jobs=None):
    '''Creates the band depth, it works for both absorption bands by changing the corresponding inputs. 
    
    Inputs:
    removed_cube = continuum-removed cube,
    minimum = the minimum image,
    wavelengths = the wavelengths.
    block_size= Size of the block during the paralelization.
    n_jobs= Amount of cores for paralelization. Use -1 for all cores, or negative to leave some cores free.
    
    Output:
    The band depth of the selected absorption band.'''

    y, z = removed_cube[0, :, :].shape
    wavelengths = wavelengths[0:74]
    cube_depth = removed_cube[0, :, :].copy()

    # Resolve number of jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_pixel(a, b):
        input_depth = removed_cube.data[:, a, b]
        if minimum[a, b] == 0:
            return 0
        else:
            input_min = minimum.data[a, b]
            pre_input_minp = np.where(wavelengths == input_min)[0][0]
            minp = int(pre_input_minp)
            band_depth_val = 1 - input_depth[minp]
            return band_depth_val

    def process_block(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                block_result[a - r_start, b - c_start] = process_pixel(a, b)
        return block_result

    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    full_array = np.zeros((y, z), dtype=np.float32)
    for idx, (r_start, r_end, c_start, c_end) in enumerate(blocks):
        full_array[r_start:r_end, c_start:c_end] = results[idx]

    cube_depth.data = full_array
    cube_depth.data[cube_depth.data == 0] = np.nan

    return cube_depth


def SSI(gauss_cube, shoulder1, wavelengths, block_size=100, n_jobs=None):
    '''Creates the spectral slope at 1 um index. This is done between the 540 nm band and the left shoulder of the 1 um band.

    Inputs: 
    gauss_cube = the filtered cube, 
    shoulder1 = the right shoulder of the 1 um band,
    wavelengths = the wavelengths.
    block_size = size of processing block for parallelization.
    n_jobs = number of parallel jobs (cores).

    Output:
    The spectral slope at 1 um.'''

    SSI = gauss_cube[0, :, :].copy()
    y, z = gauss_cube[0, :, :].shape

    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_block(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                if shoulder1[a, b] == 0:
                    block_result[a - r_start, b - c_start] = 0
                else:
                    input_SS1200 = gauss_cube.data[:, a, b]
                    input_shoulder1 = shoulder1.data[a, b]
                    pre_shoulder1 = np.where(wavelengths == input_shoulder1)[0]
                    if len(pre_shoulder1) == 0:
                        block_result[a - r_start, b - c_start] = 0
                        continue
                    shoulder1p = int(pre_shoulder1[0])
                    SS = ((input_SS1200[shoulder1p]) - input_SS1200[0]) / (((wavelengths[shoulder1p]) - 0.54084) * 0.54084)
                    block_result[a - r_start, b - c_start] = SS
        return block_result

    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    full_result = np.zeros((y, z), dtype=np.float32)
    for idx, (r_start, r_end, c_start, c_end) in enumerate(blocks):
        full_result[r_start:r_end, c_start:c_end] = results[idx]

    SSI.data = full_result
    SSI.data[SSI.data == 0] = np.nan

    return SSI


def RGB1 (gauss_cube,SSI_cube,BDI_cube,BDII_cube):
    '''Creates the RGB1 index. R: SSBI, G: BDI, B: BDII.
    
    Inputs:
    gauss_cube = the filtered cube, 
    SSI_cube = the spectral slope index,
    BDI_cube = the band depth at 1 um,
    BDII_cube = the band depth at 2 um.
    
    Output:
    The RGB1 composite.'''
    
    RGB1=gauss_cube[0:3,:,:].copy()
    RGB1.data=np.dstack((SSI_cube,BDI_cube,BDII_cube)).transpose(2,0,1)
    return RGB1


def RGB2 (gauss_cube,SSI_cube, R540_cube, BCII_cube):
    '''Creates the RGB2 index. R: SSBI, G: R540, B: BCII.
    
    Inputs:
    gauss_cube = the filtered cube, 
    SSI_cube = the spectral slope index,
    R540_cube = the reflectance at 540 nm,
    BCII_cube = the band center at 2 um.
    
    Output:
    The RGB2 composite.'''
    
    RGB2=gauss_cube[0:3,:,:].copy()
    RGB2.data=np.dstack((SSI_cube,R540_cube,BCII_cube)).transpose(2,0,1)
    return RGB2


def RGB3 (gauss_cube,SSI_cube,R540_cube,BDI_cube):
    '''Creates the RGB3 index. R: SSBI, G: R540, B: BDI.
    
    Inputs:
    gauss_cube = the filtered cube, 
    SSI_cube = the spectral slope index,
    R540_cube = the reflectance at 540 nm,
    BDI_cube = the band depth at 1 um.
    
    Output:
    The RGB3 composite.'''
    
    RGB3=gauss_cube[0:3,:,:].copy()
    RGB3.data=np.dstack((SSI_cube,R540_cube,BDI_cube)).transpose(2,0,1)
    return RGB3


def RGB6 (removed_cube):
    '''Creates the RGB6 index. R: BD950, G: BD1050, B: BD1250.
    
    Inputs:
    removed_cube = the continuum-removed cube.
    
    Output:
    The RGB6 composite.'''
    
    RGB6=removed_cube[0:3,:,:].copy()
    RGB6_R=1-removed_cube.data[16,:,:]
    RGB6_G=1-removed_cube.data[21,:,:]
    RGB6_B=1-removed_cube.data[31,:,:]
    RGB6.data=np.dstack((RGB6_R,RGB6_G,RGB6_B)).transpose(2,0,1)
    RGB6.data[RGB6.data==1]=np.nan
    return RGB6


def RGB7 (gauss_cube,IBD1000,IBD2000, R1580):
    '''Creates the RGB6 index. R: IBDI, G: IBDII, B: R1580.
    
    Inputs:
    gauss_cube = the filtered cube.
    IBD1000 = the integrated band depth at 1 um,
    IBD2000 = the integrated band depth at 2 um,
    R1580 = the reflectance at 1580.
    
    Output:
    The RGB7 composite.'''
    
    RGB7=gauss_cube[0:3,:,:].copy()
    RGB7.data=np.dstack((IBD1000,IBD2000,R1580)).transpose(2,0,1)
    return RGB7


def BA (removed_cube,wavelengths,shoulder0,shoulder1):
    '''Creates the band area index, it works for both absorption bands by changing the corresponding inputs. 
    
    Inputs: 
    removed_cube = the continuum-removed cube,
    wavelengths = the wavelengths,
    shoulder0 = the left shoulder of the band, 
    shoulder1 = the right shoulder of the band.
    
    Output:
    The band area of the selected absorption band.'''
    
    y,z=removed_cube[0,:,:].shape
    #Finding the spectral resolution, neccesary to find the area
    SR=np.diff(wavelengths) 
    #Adding the first value
    SR=np.append(39.92,SR)   

    #Calculating the band area
    BAI=removed_cube[0,:,:].copy()
    stack_BAI=[]
    for a in range(removed_cube.data.shape[1]):
        for b in range(removed_cube.data.shape[2]):
            #The shoulders limits the area calculation
            s0 = shoulder0.data[a,b]  
            s1 = shoulder1.data[a,b]
                    
            if s0 == 0 or s1 == 0:   
                    stack_BAI.append(0)
            else:
                 #Creating the range
                start = np.where(wavelengths == s0)[0][0].item() 
                end = np.where(wavelengths == s1)[0][0].item()
           
                input_SR= SR[start:end]
                input_CCA= removed_cube.data[:,a,b]
            
                sum3=0
                for c in range(start, end):
                    #Calculates the area by te sum of the spectral resolution multiplied by 1 minus the reflectance
                    sum3 += ((1 - input_CCA[c-start]) * input_SR[c-start])  
                
                stack_BAI.append(sum3)
        
    stack_BAIa=np.array(stack_BAI)
    BAI.data=stack_BAIa.reshape(y,z)
    BAI.data[BAI.data==0]=np.nan
    return BAI


def RGB4 (gauss_cube,wavelengths,shoulder0,shoulder1,minimum_1000,minimum_2000):
    '''Creates the RGB4 index. R: BCI, G: BCII, B: BAI.
    
    Inputs:
    gauss_cube = the filtered cube,
    wavelengths = the wavelengths,
    shoulder0 = the left shoulder of the band, 
    shoulder1 = the right shoulder of the band,
    minimum_1000 = the band minimum at 1 um,
    minimum_2000 = the band minimum at 2 um.
    
    Output:
    The RGB4 composite.'''
    
    y,z=gauss_cube[0,:,:].shape
    #Finding the spectral resolution, neccesary to find the area
    SR=np.diff(wavelengths)  
    #Adding the first value
    SR=np.append(39.92,SR)  

    #Calculating the band area
    BAI=gauss_cube[0,:,:].copy()
    stack_BAI=[]
    for a in range(gauss_cube.data.shape[1]):
        for b in range(gauss_cube.data.shape[2]):
            #The shoulders limits the area calculation
            s0 = shoulder0.data[a,b]  
            s1 = shoulder1.data[a,b]
                    
            if s0 == 0 or s1 == 0:   
                    stack_BAI.append(0)
            else:
                #Creating the range
                start = np.where(wavelengths == s0)[0][0].item()  
                end = np.where(wavelengths == s1)[0][0].item()
           
                input_SR= SR[start:end]
                input_CCA= gauss_cube.data[:,a,b]
            
                sum3=0
                for c in range(start, end):
                    #Calculates the area by te sum of the spectral resolution multiplied by 1 minus the reflectance
                    sum3 += ((1 - input_CCA[c-start]) * input_SR[c-start])  
                
                stack_BAI.append(sum3)
        
    stack_BAIa=np.array(stack_BAI)
    BAI.data=stack_BAIa.reshape(y,z)

    #Creating the RGB
    CCA=gauss_cube[0:3,:,:].copy()
    CCA.data=np.dstack((minimum_1000,minimum_2000,BAI)).transpose(2,0,1)
    CCA.data[CCA.data==0]=np.nan
    return CCA


def ASY(removed_cube, wavelengths, shoulder0, shoulder1, min1000, block_size=100, n_jobs=None):
    '''Creates the band asymmetry index. Works for absorption bands by specifying appropriate shoulders and band minimum.

    Inputs: 
    removed_cube = the continuum-removed cube (reflectance values from 0 to 1),
    wavelengths = array of wavelength values,
    shoulder0 = left shoulder wavelength of the band,
    shoulder1 = right shoulder wavelength of the band,
    min1000 = wavelength of the band minimum.
    block_size = block size for parallel processing.
    n_jobs = number of cores to use. -1 for all cores, negative leaves that many cores free.

    Output:
    2D image of band asymmetry values for each pixel, with optional local parabolic smoothing.
    '''

    # Compute spectral resolution
    SR = np.diff(wavelengths)
    SR = np.append(39.92, SR)  # Add the first resolution manually

    y, z = removed_cube[0, :, :].shape
    ASY = removed_cube[0, :, :].copy()

    # Configure number of parallel jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    # Define block processor
    def process_block(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                if removed_cube[39, a, b] == 0:
                    block_result[a - r_start, b - c_start] = 0
                    continue

                s00 = shoulder0.data[a, b]
                s11 = shoulder1.data[a, b]
                input_min1000 = min1000.data[a, b]

                if s00 == 0 or s11 == 0:
                    block_result[a - r_start, b - c_start] = 0
                else:
                    try:
                        start1 = np.where(wavelengths == s00)[0][0].item()
                        end1 = np.where(wavelengths == s11)[0][0].item()
                        middle = np.where(wavelengths == input_min1000)[0][0].item()

                        input_SR1 = SR[start1:middle + 1]
                        input_SR2 = SR[middle + 1:end1]
                        input_CCA = removed_cube.data[:, a, b]

                        sum4 = sum((1 - input_CCA[c]) * SR[c] for c in range(start1, middle + 1))
                        sum5 = sum((1 - input_CCA[d]) * SR[d] for d in range(middle + 1, end1))

                        total = sum4 + sum5
                        if sum4 == 0 or total == 0:
                            block_result[a - r_start, b - c_start] = 0
                        elif sum4 > sum5:
                            block_result[a - r_start, b - c_start] = -((sum4 - sum5) * 100 / total)
                        else:
                            block_result[a - r_start, b - c_start] = ((sum5 - sum4) * 100 / total)
                    except:
                        block_result[a - r_start, b - c_start] = 0
        return block_result

    # Create spatial blocks
    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    # Parallel execution of blocks
    results = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    # Assemble results
    ASY_values = np.zeros((y, z), dtype=np.float32)
    idx = 0
    for r_start, r_end, c_start, c_end in blocks:
        ASY_values[r_start:r_end, c_start:c_end] = results[idx]
        idx += 1

    # Local 2D parabolic smoothing (11x11 window)
    ASY_smoothed = ASY.data.copy()
    offset = 3
    for i in range(y):
        for j in range(z):
            if ASY_values[i, j] == 0:
                ASY_smoothed[i, j] = 0
                continue

            i_min = max(0, i - offset)
            i_max = min(y, i + offset + 1)
            j_min = max(0, j - offset)
            j_max = min(z, j + offset + 1)

            window = ASY_values[i_min:i_max, j_min:j_max].flatten()
            x = np.arange(len(window))

            if len(window) >= 6 and not np.all(np.isnan(window)):
                try:
                    fit = np.polyfit(x, window, 2)
                    smoothed_val = np.polyval(fit, x[len(x) // 2])
                    ASY_smoothed[i, j] = smoothed_val
                except:
                    ASY_smoothed[i, j] = ASY_values[i, j]
            else:
                ASY_smoothed[i, j] = ASY_values[i, j]

    # Replace 0s with NaN (optional)
    ASY_smoothed[ASY_smoothed == 0] = np.nan
    ASY.data = ASY_smoothed

    return ASY


def RGB5(gauss_cube, wavelengths, shoulder0, shoulder1, min1000, min2000):
    '''Creates the RGB5 index. R: ASY, G: BCI, B: BCII.
    
    Inputs:
    gauss_cube = the filtered cube.
    wavelengths = the wavelengths,
    shoulder0 = the left shoulder of the band, 
    shoulder1 = the right shoulder of the band, 
    min1000 = the minimum at 1 um,
    min2000 = the minimum at 2 um.
    
    Output:
    The RGB5 RGB composite.'''
    
    # Compute ASY directly using the existing ASY() function
    ASY_img = ASY(gauss_cube, wavelengths, shoulder0, shoulder1, min1000)
    
    # Create copy of first 3 bands (shape/metadata preserved)
    RGB5 = gauss_cube[0:3, :, :].copy()
    
    # Combine the ASY, min1000 and min2000 values into an RGB composite
    RGB5.data = np.dstack((ASY_img.data, min1000.data, min2000.data)).transpose(2, 0, 1)
    
    # Optional: replace zeroes with NaNs
    RGB5.data[RGB5.data == 0] = np.nan

    return RGB5


def IBDII(removed_cube):
    '''Calculates the integrated band depth around the 2 um band. Internal process. 
    
    Input:
    removed_cube = the continuum-removed cube.
    
    Output:
    Integrated band depth at 2 um.'''
    
    y2000,z2000=removed_cube[0,:,:].shape
    IBDII=removed_cube[0,:,:].copy()
    #Defines the section to iterate around 2000 nm
    IBDII_slice=removed_cube[49:70,:,:]  
    stack_IBDII=[]
    
    for a in range(IBDII_slice.data.shape[1]):
        for b in range(IBDII_slice.data.shape[2]):
            sum1=0
            if IBDII_slice[19,a,b]==0: 
                    stack_IBDII.append(0)
            else:
                for c in range(IBDII_slice.data.shape[0]):
                    input_removed=IBDII_slice.data[:,a,b]
                    #Summatory
                    sum1 += (1- input_removed[c]) 
                stack_IBDII.append(sum1)
        
    stack_IBDIIa=np.array(stack_IBDII)
    IBDII.data=stack_IBDIIa.reshape(y2000,z2000)
    IBDII.data[IBDII.data==0]=np.nan
    return IBDII
    
    
def IBDI(removed_cube):
    '''Calculates the integrated band depth around the 1 um band. Internal process. 
    
    Input:
    removed_cube = the continuum-removed cube.
    
    Output:
    Integrated band depth at 1 um.'''
    
    y2000,z2000=removed_cube[0,:,:].shape
    IBDI=removed_cube[0,:,:].copy()
    #Defines the section to iterate around 1000 nm
    IBDI_slice=removed_cube[8:34,:,:]  
    stack_IBDI=[]
    
    for a in range(IBDI_slice.data.shape[1]):
        for b in range(IBDI_slice.data.shape[2]):
            sum1=0
            if IBDI_slice[19,a,b]==0: 
                    stack_IBDI.append(0)
            else:
                for c in range(IBDI_slice.data.shape[0]):
                    input_removed=IBDI_slice.data[:,a,b]
                    #Summatory
                    sum1 += (1- input_removed[c])  
                stack_IBDI.append(sum1)
        
    stack_IBDIa=np.array(stack_IBDI)
    IBDI.data=stack_IBDIa.reshape(y2000,z2000)
    IBDI.data[IBDI.data==0]=np.nan
    return IBDI


def RGB8(gauss_cube, removed_cube, block_size=50, n_jobs=None):
    '''Creates the RGB8 index. R: BD1900, G: IBD2000, B: IBD1000.
    
    Inputs:
    gauss_cube = the filtered cube.
    removed_cube = continuum-removed cube.
    block_size = size of processing block for parallelization.
    n_jobs = number of parallel jobs (cores).
    
    Output:
    The RGB8 RGB composite.'''

    y, z = removed_cube[0, :, :].shape

    # Band 1: Band depth at 1900 nm
    RGB81 = 1 - removed_cube[55, :, :]
    RGB81.data[RGB81.data == 1] = np.nan

    # Prepare containers for bands 2 and 3
    RGB82 = removed_cube[0, :, :].copy()
    RGB83 = removed_cube[0, :, :].copy()

    # Slices for bands 2 and 3
    RGB82_slice = removed_cube[49:70, :, :]
    RGB83_slice = removed_cube[8:34, :, :]

    # Resolve n_jobs
    if n_jobs is None:
        n_jobs = 1
    elif n_jobs < 0:
        n_jobs = max(1, multiprocessing.cpu_count() + n_jobs)

    def process_block_82(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                if RGB82_slice[19, a, b] == 0:
                    block_result[a - r_start, b - c_start] = 0
                else:
                    input_removed = RGB82_slice.data[:, a, b]
                    block_result[a - r_start, b - c_start] = np.sum(1 - input_removed)
        return block_result

    def process_block_83(r_start, r_end, c_start, c_end):
        block_result = np.zeros((r_end - r_start, c_end - c_start), dtype=np.float32)
        for a in range(r_start, r_end):
            for b in range(c_start, c_end):
                if RGB83_slice[19, a, b] == 0:
                    block_result[a - r_start, b - c_start] = 0
                else:
                    input_removed4 = RGB83_slice.data[:, a, b]
                    block_result[a - r_start, b - c_start] = np.sum(1 - input_removed4)
        return block_result

    blocks = []
    for r in range(0, y, block_size):
        for c in range(0, z, block_size):
            r_end = min(r + block_size, y)
            c_end = min(c + block_size, z)
            blocks.append((r, r_end, c, c_end))

    results_82 = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block_82)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )
    results_83 = Parallel(n_jobs=n_jobs, prefer="processes")(
        delayed(process_block_83)(r_start, r_end, c_start, c_end) for (r_start, r_end, c_start, c_end) in blocks
    )

    full_82 = np.zeros((y, z), dtype=np.float32)
    full_83 = np.zeros((y, z), dtype=np.float32)

    for idx, (r_start, r_end, c_start, c_end) in enumerate(blocks):
        full_82[r_start:r_end, c_start:c_end] = results_82[idx]
        full_83[r_start:r_end, c_start:c_end] = results_83[idx]

    RGB82.data = full_82
    RGB83.data = full_83

    # Compose RGB cube
    RGB8_total = gauss_cube[0:3, :, :].copy()
    RGB8_total.data = np.dstack((RGB81, RGB82, RGB83)).transpose(2, 0, 1)
    RGB8_total.data[RGB8_total.data == 0] = np.nan

    return RGB8_total


import numpy as np
from specutils import Spectrum
from specutils.manipulation import gaussian_smooth
import astropy.units as u
import cv2

def fourier_filter(original_cube,percentage_width,percentage_height):
    '''Performs the fourier filtering of the cube in the spatial domain. 
    
    Inputs:
    original_cube = the prepared cube, 
    percentage_width = width of the filter in percentange (60 recommended),
    percentage_height = height of the filter in percentage (2 recommended).
    
    Outputs:
    The cube after the fourier filter.'''

    #Check input percentage values
    if percentage_width > 100 or percentage_height > 100:
        raise ValueError("Invalid percentage")
    #Creating a new cube to put the final data
    fourier_cube=original_cube.copy()  
    #Getting the dimensions to do the mask
    rows,cols=original_cube[0,:,:].shape  
    mask= np.ones((rows, cols, 2), np.uint8)
    y,z=original_cube[0,:,:].shape
    #Calculating the input size of the box
    filter_width=int((z*((100-percentage_width)/100))/2) 
    filter_height=int((y*(percentage_height/100))/2)
    #Horizontal size of the fitler
    cv2.rectangle(mask, (0,((y//2)-filter_height)), (((z//2)-filter_width),((y//2)+filter_height)), 0, -1)  
    #Vertical size of the filter
    cv2.rectangle(mask, (((z//2)+filter_width),((y//2)-filter_height)), (z,((y//2)+filter_height)), 0, -1)  
    #Creating a list to store the values
    stack_fourier=[]  

    for band in range(original_cube.data.shape[0]):
        #Fourier transform 
        input_img=original_cube[band,:,:].data
        fouraster=cv2.dft(input_img, flags=cv2.DFT_COMPLEX_OUTPUT)  
        fouraster_shift=np.fft.fftshift(fouraster)
        #Mask used to fitler the noise
        mfouraster=fouraster_shift*mask  
        #Inverse fourirer to recover the image
        m_ishift= np.fft.ifftshift(mfouraster)  
        fourier_raster=cv2.idft(m_ishift)/(y*z)
        fourier_raster= cv2.magnitude(fourier_raster[:, :, 0], fourier_raster[:, :, 1])
        fourier_raster2=fourier_raster.astype(np.float32)
        stack_fourier.append(fourier_raster2)
        
    fourier_cube.data=np.array(stack_fourier)
    
    #Creating a mask to avoid the non-data regions
    mask_cube=original_cube.copy()
    mask_cube.data[mask_cube.data != 0]=1
    fourier_cube_final=fourier_cube*mask_cube
    return fourier_cube_final


def gauss_filter (cube_filtered,wavelengths):
    '''Performs the Gaussian filtering in the spectral domain. 
    
    Inputs:
    cube_filtered = fourier-filtered cube,
    wavelengths = the wavelengths file.
    
    Outputs:
    The filtered cube ready for the continuum removal.'''

    #Transposing data to ingest in Spec1d2
    tcube=np.transpose(cube_filtered.data,(2,1,0))  
    #Transforming data to spec1d2 file type 
    spec1d2=Spectrum(spectral_axis=wavelengths[0::]*u.AA, flux=tcube*u.dimensionless_unscaled)  
    #Filtering with a 1d gaussian function, using a 4 pixel kernel
    spec1d2_gsmooth=gaussian_smooth(spec1d2, stddev=1) 
    #Slicing the filtered cube to avoid the affected data at longer wavelengths an re-attaching the original
    tcube_smooth=spec1d2_gsmooth.flux.value  
    tslice1=tcube_smooth[:,:,0:79]
    tslice2=tcube[:,:,79:83]
    spec1d2_filtered=np.dstack((tslice1,tslice2))
    #Saving the filtered data in a new cube, copied from the original to maintain the projection
    M3_gaussfilter=cube_filtered.copy()  
    M3_gaussfilter.data=np.transpose(spec1d2_filtered,(2,1,0))
    
    #Creating a mask to avoid no-data
    mask_cube=cube_filtered.copy()
    mask_cube.data[mask_cube.data != 0]=1
    gauss_cube_final=M3_gaussfilter*mask_cube
    return(gauss_cube_final)


from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description =f.read()

setup(
    name='MoonIndex',
    version='3.0.0',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Javier Suarez',
    author_email='jsuarezvalencia@constructor.university',
    license='GPL',
    install_requires=['astropy>=5.3.4'
    , 'jmespath>=1.0.1'                  
    , 'matplotlib>=3.8.0'
    , 'numpy>=1.26.1'
    , 'opencv-python>=4.8.1.78'
    , 'pysptools>=0.15.0'
    , 'python-dateutil>=2.8.2'
    , 'rasterio>=1.3.9'
    , 'rioxarray>=0.15.0'
    , 'scikit-learn>=1.3.2'
    , 'scipy>=1.11.3'
    , 'specutils>=1.12.0'
    , 'xarray>=2023.10.1'
                     ],
    python_requires='>=3.10'
)